@@ -1,19 +1,20 @@
-// Copyright (C) 2014 The Protocol Authors.
-
-//go:generate -command counterfeiter go run github.com/maxbrunsfeld/counterfeiter/v6
+// Copyright (C) 2014 The Syncthing Authors.
+//
+// This Source Code Form is subject to the terms of the Mozilla Public
+// License, v. 2.0. If a copy of the MPL was not distributed with this file,
+// You can obtain one at https://mozilla.org/MPL/2.0/.
 
 // Prevents import loop, for internal testing
-//go:generate counterfeiter -o mocked_connection_info_test.go --fake-name mockedConnectionInfo . ConnectionInfo
+//go:generate go tool counterfeiter -o mocked_connection_info_test.go --fake-name mockedConnectionInfo . ConnectionInfo
 //go:generate go run ../../script/prune_mocks.go -t mocked_connection_info_test.go
 
-//go:generate counterfeiter -o mocks/connection_info.go --fake-name ConnectionInfo . ConnectionInfo
-//go:generate counterfeiter -o mocks/connection.go --fake-name Connection . Connection
+//go:generate go tool counterfeiter -o mocks/connection_info.go --fake-name ConnectionInfo . ConnectionInfo
+//go:generate go tool counterfeiter -o mocks/connection.go --fake-name Connection . Connection
 
 package protocol
 
 import (
 	"context"
-	"crypto/sha256"
 	"encoding/binary"
 	"errors"
 	"fmt"
@@ -25,6 +26,10 @@ import (
 	"time"
 
 	lz4 "github.com/pierrec/lz4/v4"
+	"google.golang.org/protobuf/proto"
+
+	"github.com/syncthing/syncthing/internal/gen/bep"
+	"github.com/syncthing/syncthing/internal/protoutil"
 )
 
 const (
@@ -46,72 +51,20 @@ const (
 
 	// DesiredPerFileBlocks is the number of blocks we aim for per file
 	DesiredPerFileBlocks = 2000
+
+	SyntheticDirectorySize = 128
+
+	// don't bother compressing messages smaller than this many bytes
+	compressionThreshold = 128
 )
 
-// BlockSizes is the list of valid block sizes, from min to max
-var BlockSizes []int
-
-// For each block size, the hash of a block of all zeroes
-var sha256OfEmptyBlock = map[int][sha256.Size]byte{
-	128 << KiB: {0xfa, 0x43, 0x23, 0x9b, 0xce, 0xe7, 0xb9, 0x7c, 0xa6, 0x2f, 0x0, 0x7c, 0xc6, 0x84, 0x87, 0x56, 0xa, 0x39, 0xe1, 0x9f, 0x74, 0xf3, 0xdd, 0xe7, 0x48, 0x6d, 0xb3, 0xf9, 0x8d, 0xf8, 0xe4, 0x71},
-	256 << KiB: {0x8a, 0x39, 0xd2, 0xab, 0xd3, 0x99, 0x9a, 0xb7, 0x3c, 0x34, 0xdb, 0x24, 0x76, 0x84, 0x9c, 0xdd, 0xf3, 0x3, 0xce, 0x38, 0x9b, 0x35, 0x82, 0x68, 0x50, 0xf9, 0xa7, 0x0, 0x58, 0x9b, 0x4a, 0x90},
-	512 << KiB: {0x7, 0x85, 0x4d, 0x2f, 0xef, 0x29, 0x7a, 0x6, 0xba, 0x81, 0x68, 0x5e, 0x66, 0xc, 0x33, 0x2d, 0xe3, 0x6d, 0x5d, 0x18, 0xd5, 0x46, 0x92, 0x7d, 0x30, 0xda, 0xad, 0x6d, 0x7f, 0xda, 0x15, 0x41},
-	1 << MiB:   {0x30, 0xe1, 0x49, 0x55, 0xeb, 0xf1, 0x35, 0x22, 0x66, 0xdc, 0x2f, 0xf8, 0x6, 0x7e, 0x68, 0x10, 0x46, 0x7, 0xe7, 0x50, 0xab, 0xb9, 0xd3, 0xb3, 0x65, 0x82, 0xb8, 0xaf, 0x90, 0x9f, 0xcb, 0x58},
-	2 << MiB:   {0x56, 0x47, 0xf0, 0x5e, 0xc1, 0x89, 0x58, 0x94, 0x7d, 0x32, 0x87, 0x4e, 0xeb, 0x78, 0x8f, 0xa3, 0x96, 0xa0, 0x5d, 0xb, 0xab, 0x7c, 0x1b, 0x71, 0xf1, 0x12, 0xce, 0xb7, 0xe9, 0xb3, 0x1e, 0xee},
-	4 << MiB:   {0xbb, 0x9f, 0x8d, 0xf6, 0x14, 0x74, 0xd2, 0x5e, 0x71, 0xfa, 0x0, 0x72, 0x23, 0x18, 0xcd, 0x38, 0x73, 0x96, 0xca, 0x17, 0x36, 0x60, 0x5e, 0x12, 0x48, 0x82, 0x1c, 0xc0, 0xde, 0x3d, 0x3a, 0xf8},
-	8 << MiB:   {0x2d, 0xae, 0xb1, 0xf3, 0x60, 0x95, 0xb4, 0x4b, 0x31, 0x84, 0x10, 0xb3, 0xf4, 0xe8, 0xb5, 0xd9, 0x89, 0xdc, 0xc7, 0xbb, 0x2, 0x3d, 0x14, 0x26, 0xc4, 0x92, 0xda, 0xb0, 0xa3, 0x5, 0x3e, 0x74},
-	16 << MiB:  {0x8, 0xa, 0xcf, 0x35, 0xa5, 0x7, 0xac, 0x98, 0x49, 0xcf, 0xcb, 0xa4, 0x7d, 0xc2, 0xad, 0x83, 0xe0, 0x1b, 0x75, 0x66, 0x3a, 0x51, 0x62, 0x79, 0xc8, 0xb9, 0xd2, 0x43, 0xb7, 0x19, 0x64, 0x3e},
-}
-
-var (
-	errNotCompressible = errors.New("not compressible")
-)
-
-func init() {
-	for blockSize := MinBlockSize; blockSize <= MaxBlockSize; blockSize *= 2 {
-		BlockSizes = append(BlockSizes, blockSize)
-		if _, ok := sha256OfEmptyBlock[blockSize]; !ok {
-			panic("missing hard coded value for sha256 of empty block")
-		}
-	}
-	BufferPool = newBufferPool()
-}
-
-// BlockSize returns the block size to use for the given file size
-func BlockSize(fileSize int64) int {
-	var blockSize int
-	for _, blockSize = range BlockSizes {
-		if fileSize < DesiredPerFileBlocks*int64(blockSize) {
-			break
-		}
-	}
-
-	return blockSize
-}
+var errNotCompressible = errors.New("not compressible")
 
 const (
 	stateInitial = iota
 	stateReady
 )
 
-// FileInfo.LocalFlags flags
-const (
-	FlagLocalUnsupported = 1 << 0 // The kind is unsupported, e.g. symlinks on Windows
-	FlagLocalIgnored     = 1 << 1 // Matches local ignore patterns
-	FlagLocalMustRescan  = 1 << 2 // Doesn't match content on disk, must be rechecked fully
-	FlagLocalReceiveOnly = 1 << 3 // Change detected on receive only folder
-
-	// Flags that should result in the Invalid bit on outgoing updates
-	LocalInvalidFlags = FlagLocalUnsupported | FlagLocalIgnored | FlagLocalMustRescan | FlagLocalReceiveOnly
-
-	// Flags that should result in a file being in conflict with its
-	// successor, due to us not having an up to date picture of its state on
-	// disk.
-	LocalConflictFlags = FlagLocalUnsupported | FlagLocalIgnored | FlagLocalReceiveOnly
-
-	LocalAllFlags = FlagLocalUnsupported | FlagLocalIgnored | FlagLocalMustRescan | FlagLocalReceiveOnly
-)
-
 var (
 	ErrClosed             = errors.New("connection closed")
 	ErrTimeout            = errors.New("read timeout")
@@ -125,17 +78,28 @@ var (
 
 type Model interface {
 	// An index was received from the peer device
-	Index(deviceID DeviceID, folder string, files []FileInfo) error
+	Index(conn Connection, idx *Index) error
 	// An index update was received from the peer device
-	IndexUpdate(deviceID DeviceID, folder string, files []FileInfo) error
+	IndexUpdate(conn Connection, idxUp *IndexUpdate) error
 	// A request was made by the peer device
-	Request(deviceID DeviceID, folder, name string, blockNo, size int32, offset int64, hash []byte, weakHash uint32, fromTemporary bool) (RequestResponse, error)
+	Request(conn Connection, req *Request) (RequestResponse, error)
 	// A cluster configuration message was received
-	ClusterConfig(deviceID DeviceID, config ClusterConfig) error
+	ClusterConfig(conn Connection, config *ClusterConfig) error
 	// The peer device closed the connection or an error occurred
-	Closed(device DeviceID, err error)
+	Closed(conn Connection, err error)
 	// The peer device sent progress updates for the files it is currently downloading
-	DownloadProgress(deviceID DeviceID, folder string, updates []FileDownloadProgressUpdate) error
+	DownloadProgress(conn Connection, p *DownloadProgress) error
+}
+
+// rawModel is the Model interface, but without the initial Connection
+// parameter. Internal use only.
+type rawModel interface {
+	Index(*Index) error
+	IndexUpdate(*IndexUpdate) error
+	Request(*Request) (RequestResponse, error)
+	ClusterConfig(*ClusterConfig) error
+	Closed(err error)
+	DownloadProgress(*DownloadProgress) error
 }
 
 type RequestResponse interface {
@@ -145,17 +109,39 @@ type RequestResponse interface {
 }
 
 type Connection interface {
+	// Send an Index message to the peer device. The message in the
+	// parameter may be altered by the connection and should not be used
+	// further by the caller.
+	Index(ctx context.Context, idx *Index) error
+
+	// Send an Index Update message to the peer device. The message in the
+	// parameter may be altered by the connection and should not be used
+	// further by the caller.
+	IndexUpdate(ctx context.Context, idxUp *IndexUpdate) error
+
+	// Send a Request message to the peer device. The message in the
+	// parameter may be altered by the connection and should not be used
+	// further by the caller.
+	Request(ctx context.Context, req *Request) ([]byte, error)
+
+	// Send a Cluster Configuration message to the peer device. The message
+	// in the parameter may be altered by the connection and should not be
+	// used further by the caller.
+	// For any folder that must be encrypted for the connected device, the
+	// password must be provided.
+	ClusterConfig(config *ClusterConfig, passwords map[string]string)
+
+	// Send a Download Progress message to the peer device. The message in
+	// the parameter may be altered by the connection and should not be used
+	// further by the caller.
+	DownloadProgress(ctx context.Context, dp *DownloadProgress)
+
 	Start()
-	SetFolderPasswords(passwords map[string]string)
 	Close(err error)
-	ID() DeviceID
-	Index(ctx context.Context, folder string, files []FileInfo) error
-	IndexUpdate(ctx context.Context, folder string, files []FileInfo) error
-	Request(ctx context.Context, folder string, name string, blockNo int, offset int64, size int, hash []byte, weakHash uint32, fromTemporary bool) ([]byte, error)
-	ClusterConfig(config ClusterConfig)
-	DownloadProgress(ctx context.Context, folder string, updates []FileDownloadProgressUpdate)
+	DeviceID() DeviceID
 	Statistics() Statistics
 	Closed() <-chan struct{}
+
 	ConnectionInfo
 }
 
@@ -168,14 +154,17 @@ type ConnectionInfo interface {
 	String() string
 	Crypto() string
 	EstablishedAt() time.Time
+	ConnectionID() string
 }
 
 type rawConnection struct {
 	ConnectionInfo
 
-	id        DeviceID
-	receiver  Model
+	deviceID  DeviceID
+	idString  string
+	model     rawModel
 	startTime time.Time
+	started   chan struct{}
 
 	cr     *countingReader
 	cw     *countingWriter
@@ -187,7 +176,7 @@ type rawConnection struct {
 
 	idxMut sync.Mutex // ensures serialization of Index calls
 
-	inbox                 chan message
+	inbox                 chan proto.Message
 	outbox                chan asyncMessage
 	closeBox              chan asyncMessage
 	clusterConfigBox      chan *ClusterConfig
@@ -196,6 +185,7 @@ type rawConnection struct {
 	closeOnce             sync.Once
 	sendCloseOnce         sync.Once
 	compression           Compression
+	startStopMut          sync.Mutex // start and stop must be serialized
 
 	loopWG sync.WaitGroup // Need to ensure no leftover routines in testing
 }
@@ -205,15 +195,8 @@ type asyncResult struct {
 	err error
 }
 
-type message interface {
-	ProtoSize() int
-	Marshal() ([]byte, error)
-	MarshalTo([]byte) (int, error)
-	Unmarshal([]byte) error
-}
-
 type asyncMessage struct {
-	msg  message
+	msg  proto.Message
 	done chan struct{} // done closes when we're done sending the message
 }
 
@@ -231,34 +214,45 @@ const (
 // Should not be modified in production code, just for testing.
 var CloseTimeout = 10 * time.Second
 
-func NewConnection(deviceID DeviceID, reader io.Reader, writer io.Writer, closer io.Closer, receiver Model, connInfo ConnectionInfo, compress Compression, passwords map[string]string) Connection {
+func NewConnection(deviceID DeviceID, reader io.Reader, writer io.Writer, closer io.Closer, model Model, connInfo ConnectionInfo, compress Compression, keyGen *KeyGenerator) Connection {
+	// We create the wrapper for the model first, as it needs to be passed
+	// in at the lowest level in the stack. At the end of construction,
+	// before returning, we add the connection to cwm so that it can be used
+	// by the model.
+	cwm := &connectionWrappingModel{model: model}
+
 	// Encryption / decryption is first (outermost) before conversion to
 	// native path formats.
-	nm := makeNative(receiver)
-	em := &encryptedModel{model: nm, folderKeys: newFolderKeyRegistry(passwords)}
+	nm := makeNative(cwm)
+	em := newEncryptedModel(nm, keyGen)
 
 	// We do the wire format conversion first (outermost) so that the
 	// metadata is in wire format when it reaches the encryption step.
 	rc := newRawConnection(deviceID, reader, writer, closer, em, connInfo, compress)
-	ec := encryptedConnection{ConnectionInfo: rc, conn: rc, folderKeys: em.folderKeys}
+	ec := newEncryptedConnection(rc, rc, em.folderKeys, keyGen)
 	wc := wireFormatConnection{ec}
 
+	cwm.conn = wc
 	return wc
 }
 
-func newRawConnection(deviceID DeviceID, reader io.Reader, writer io.Writer, closer io.Closer, receiver Model, connInfo ConnectionInfo, compress Compression) *rawConnection {
-	cr := &countingReader{Reader: reader}
-	cw := &countingWriter{Writer: writer}
+func newRawConnection(deviceID DeviceID, reader io.Reader, writer io.Writer, closer io.Closer, receiver rawModel, connInfo ConnectionInfo, compress Compression) *rawConnection {
+	idString := deviceID.String()
+	cr := &countingReader{Reader: reader, idString: idString}
+	cw := &countingWriter{Writer: writer, idString: idString}
+	registerDeviceMetrics(idString)
 
 	return &rawConnection{
 		ConnectionInfo:        connInfo,
-		id:                    deviceID,
-		receiver:              receiver,
+		deviceID:              deviceID,
+		idString:              deviceID.String(),
+		model:                 receiver,
+		started:               make(chan struct{}),
 		cr:                    cr,
 		cw:                    cw,
 		closer:                closer,
 		awaiting:              make(map[int]chan asyncResult),
-		inbox:                 make(chan message),
+		inbox:                 make(chan proto.Message),
 		outbox:                make(chan asyncMessage),
 		closeBox:              make(chan asyncMessage),
 		clusterConfigBox:      make(chan *ClusterConfig),
@@ -270,8 +264,22 @@ func newRawConnection(deviceID DeviceID, reader io.Reader, writer io.Writer, clo
 }
 
 // Start creates the goroutines for sending and receiving of messages. It must
-// be called exactly once after creating a connection.
+// be called once after creating a connection. It should only be called once,
+// subsequent calls will have no effect.
 func (c *rawConnection) Start() {
+	c.startStopMut.Lock()
+	defer c.startStopMut.Unlock()
+
+	select {
+	case <-c.started:
+		return
+	case <-c.closed:
+		// we have already closed the connection before starting processing
+		// on it.
+		return
+	default:
+	}
+
 	c.loopWG.Add(5)
 	go func() {
 		c.readerLoop()
@@ -294,47 +302,55 @@ func (c *rawConnection) Start() {
 		c.pingReceiver()
 		c.loopWG.Done()
 	}()
+
 	c.startTime = time.Now().Truncate(time.Second)
+	close(c.started)
 }
 
-func (c *rawConnection) ID() DeviceID {
-	return c.id
+func (c *rawConnection) DeviceID() DeviceID {
+	return c.deviceID
 }
 
 // Index writes the list of file information to the connected peer device
-func (c *rawConnection) Index(ctx context.Context, folder string, idx []FileInfo) error {
+func (c *rawConnection) Index(ctx context.Context, idx *Index) error {
 	select {
 	case <-c.closed:
 		return ErrClosed
+	case <-ctx.Done():
+		return ctx.Err()
 	default:
 	}
 	c.idxMut.Lock()
-	c.send(ctx, &Index{
-		Folder: folder,
-		Files:  idx,
-	}, nil)
+	c.send(ctx, idx.toWire(), nil)
 	c.idxMut.Unlock()
 	return nil
 }
 
 // IndexUpdate writes the list of file information to the connected peer device as an update
-func (c *rawConnection) IndexUpdate(ctx context.Context, folder string, idx []FileInfo) error {
+func (c *rawConnection) IndexUpdate(ctx context.Context, idxUp *IndexUpdate) error {
 	select {
 	case <-c.closed:
 		return ErrClosed
+	case <-ctx.Done():
+		return ctx.Err()
 	default:
 	}
 	c.idxMut.Lock()
-	c.send(ctx, &IndexUpdate{
-		Folder: folder,
-		Files:  idx,
-	}, nil)
+	c.send(ctx, idxUp.toWire(), nil)
 	c.idxMut.Unlock()
 	return nil
 }
 
 // Request returns the bytes for the specified block after fetching them from the connected peer.
-func (c *rawConnection) Request(ctx context.Context, folder string, name string, blockNo int, offset int64, size int, hash []byte, weakHash uint32, fromTemporary bool) ([]byte, error) {
+func (c *rawConnection) Request(ctx context.Context, req *Request) ([]byte, error) {
+	select {
+	case <-c.closed:
+		return nil, ErrClosed
+	case <-ctx.Done():
+		return nil, ctx.Err()
+	default:
+	}
+
 	rc := make(chan asyncResult, 1)
 
 	c.awaitingMut.Lock()
@@ -347,17 +363,8 @@ func (c *rawConnection) Request(ctx context.Context, folder string, name string,
 	c.awaiting[id] = rc
 	c.awaitingMut.Unlock()
 
-	ok := c.send(ctx, &Request{
-		ID:            id,
-		Folder:        folder,
-		Name:          name,
-		Offset:        offset,
-		Size:          size,
-		BlockNo:       blockNo,
-		Hash:          hash,
-		WeakHash:      weakHash,
-		FromTemporary: fromTemporary,
-	}, nil)
+	req.ID = id
+	ok := c.send(ctx, req.toWire(), nil)
 	if !ok {
 		return nil, ErrClosed
 	}
@@ -374,9 +381,9 @@ func (c *rawConnection) Request(ctx context.Context, folder string, name string,
 }
 
 // ClusterConfig sends the cluster configuration message to the peer.
-func (c *rawConnection) ClusterConfig(config ClusterConfig) {
+func (c *rawConnection) ClusterConfig(config *ClusterConfig, _ map[string]string) {
 	select {
-	case c.clusterConfigBox <- &config:
+	case c.clusterConfigBox <- config:
 	case <-c.closed:
 	}
 }
@@ -386,15 +393,12 @@ func (c *rawConnection) Closed() <-chan struct{} {
 }
 
 // DownloadProgress sends the progress updates for the files that are currently being downloaded.
-func (c *rawConnection) DownloadProgress(ctx context.Context, folder string, updates []FileDownloadProgressUpdate) {
-	c.send(ctx, &DownloadProgress{
-		Folder:  folder,
-		Updates: updates,
-	}, nil)
+func (c *rawConnection) DownloadProgress(ctx context.Context, dp *DownloadProgress) {
+	c.send(ctx, dp.toWire(), nil)
 }
 
 func (c *rawConnection) ping() bool {
-	return c.send(context.Background(), &Ping{}, nil)
+	return c.send(context.Background(), &bep.Ping{}, nil)
 }
 
 func (c *rawConnection) readerLoop() {
@@ -414,21 +418,27 @@ func (c *rawConnection) readerLoop() {
 		case <-c.closed:
 			return
 		}
-
 	}
 }
 
 func (c *rawConnection) dispatcherLoop() (err error) {
 	defer close(c.dispatcherLoopStopped)
-	var msg message
+	var msg proto.Message
 	state := stateInitial
 	for {
+		select {
+		case <-c.closed:
+			return ErrClosed
+		default:
+		}
 		select {
 		case msg = <-c.inbox:
 		case <-c.closed:
 			return ErrClosed
 		}
 
+		metricDeviceRecvMessages.WithLabelValues(c.idString).Inc()
+
 		msgContext, err := messageContext(msg)
 		if err != nil {
 			return fmt.Errorf("protocol error: %w", err)
@@ -436,11 +446,11 @@ func (c *rawConnection) dispatcherLoop() (err error) {
 		l.Debugf("handle %v message", msgContext)
 
 		switch msg := msg.(type) {
-		case *ClusterConfig:
+		case *bep.ClusterConfig:
 			if state == stateInitial {
 				state = stateReady
 			}
-		case *Close:
+		case *bep.Close:
 			return fmt.Errorf("closed by remote: %v", msg.Reason)
 		default:
 			if state != stateReady {
@@ -449,13 +459,7 @@ func (c *rawConnection) dispatcherLoop() (err error) {
 		}
 
 		switch msg := msg.(type) {
-		case *Index:
-			err = checkIndexConsistency(msg.Files)
-
-		case *IndexUpdate:
-			err = checkIndexConsistency(msg.Files)
-
-		case *Request:
+		case *bep.Request:
 			err = checkFilename(msg.Name)
 		}
 		if err != nil {
@@ -463,23 +467,31 @@ func (c *rawConnection) dispatcherLoop() (err error) {
 		}
 
 		switch msg := msg.(type) {
-		case *ClusterConfig:
-			err = c.receiver.ClusterConfig(c.id, *msg)
+		case *bep.ClusterConfig:
+			err = c.model.ClusterConfig(clusterConfigFromWire(msg))
 
-		case *Index:
-			err = c.handleIndex(*msg)
+		case *bep.Index:
+			idx := indexFromWire(msg)
+			if err := checkIndexConsistency(idx.Files); err != nil {
+				return newProtocolError(err, msgContext)
+			}
+			err = c.handleIndex(idx)
 
-		case *IndexUpdate:
-			err = c.handleIndexUpdate(*msg)
+		case *bep.IndexUpdate:
+			idxUp := indexUpdateFromWire(msg)
+			if err := checkIndexConsistency(idxUp.Files); err != nil {
+				return newProtocolError(err, msgContext)
+			}
+			err = c.handleIndexUpdate(idxUp)
 
-		case *Request:
-			go c.handleRequest(*msg)
+		case *bep.Request:
+			go c.handleRequest(requestFromWire(msg))
 
-		case *Response:
-			c.handleResponse(*msg)
+		case *bep.Response:
+			c.handleResponse(responseFromWire(msg))
 
-		case *DownloadProgress:
-			err = c.receiver.DownloadProgress(c.id, msg.Folder, msg.Updates)
+		case *bep.DownloadProgress:
+			err = c.model.DownloadProgress(downloadProgressFromWire(msg))
 		}
 		if err != nil {
 			return newHandleError(err, msgContext)
@@ -487,7 +499,7 @@ func (c *rawConnection) dispatcherLoop() (err error) {
 	}
 }
 
-func (c *rawConnection) readMessage(fourByteBuf []byte) (message, error) {
+func (c *rawConnection) readMessage(fourByteBuf []byte) (proto.Message, error) {
 	hdr, err := c.readHeader(fourByteBuf)
 	if err != nil {
 		return nil, err
@@ -496,7 +508,7 @@ func (c *rawConnection) readMessage(fourByteBuf []byte) (message, error) {
 	return c.readMessageAfterHeader(hdr, fourByteBuf)
 }
 
-func (c *rawConnection) readMessageAfterHeader(hdr Header, fourByteBuf []byte) (message, error) {
+func (c *rawConnection) readMessageAfterHeader(hdr *bep.Header, fourByteBuf []byte) (proto.Message, error) {
 	// First comes a 4 byte message length
 
 	if _, err := io.ReadFull(c.cr, fourByteBuf[:4]); err != nil {
@@ -512,20 +524,20 @@ func (c *rawConnection) readMessageAfterHeader(hdr Header, fourByteBuf []byte) (
 	// Then comes the message
 
 	buf := BufferPool.Get(int(msgLen))
+	defer BufferPool.Put(buf)
+
 	if _, err := io.ReadFull(c.cr, buf); err != nil {
-		BufferPool.Put(buf)
 		return nil, fmt.Errorf("reading message: %w", err)
 	}
 
 	// ... which might be compressed
 
 	switch hdr.Compression {
-	case MessageCompressionNone:
+	case bep.MessageCompression_MESSAGE_COMPRESSION_NONE:
 		// Nothing
 
-	case MessageCompressionLZ4:
+	case bep.MessageCompression_MESSAGE_COMPRESSION_LZ4:
 		decomp, err := lz4Decompress(buf)
-		BufferPool.Put(buf)
 		if err != nil {
 			return nil, fmt.Errorf("decompressing message: %w", err)
 		}
@@ -537,57 +549,58 @@ func (c *rawConnection) readMessageAfterHeader(hdr Header, fourByteBuf []byte) (
 
 	// ... and is then unmarshalled
 
+	metricDeviceRecvDecompressedBytes.WithLabelValues(c.idString).Add(float64(4 + len(buf)))
+
 	msg, err := newMessage(hdr.Type)
 	if err != nil {
-		BufferPool.Put(buf)
 		return nil, err
 	}
-	if err := msg.Unmarshal(buf); err != nil {
-		BufferPool.Put(buf)
+	if err := proto.Unmarshal(buf, msg); err != nil {
 		return nil, fmt.Errorf("unmarshalling message: %w", err)
 	}
-	BufferPool.Put(buf)
 
 	return msg, nil
 }
 
-func (c *rawConnection) readHeader(fourByteBuf []byte) (Header, error) {
+func (c *rawConnection) readHeader(fourByteBuf []byte) (*bep.Header, error) {
 	// First comes a 2 byte header length
 
 	if _, err := io.ReadFull(c.cr, fourByteBuf[:2]); err != nil {
-		return Header{}, fmt.Errorf("reading length: %w", err)
+		return nil, fmt.Errorf("reading length: %w", err)
 	}
 	hdrLen := int16(binary.BigEndian.Uint16(fourByteBuf))
 	if hdrLen < 0 {
-		return Header{}, fmt.Errorf("negative header length %d", hdrLen)
+		return nil, fmt.Errorf("negative header length %d", hdrLen)
 	}
 
 	// Then comes the header
 
 	buf := BufferPool.Get(int(hdrLen))
+	defer BufferPool.Put(buf)
+
 	if _, err := io.ReadFull(c.cr, buf); err != nil {
-		BufferPool.Put(buf)
-		return Header{}, fmt.Errorf("reading header: %w", err)
+		return nil, fmt.Errorf("reading header: %w", err)
 	}
 
-	var hdr Header
-	err := hdr.Unmarshal(buf)
-	BufferPool.Put(buf)
+	var hdr bep.Header
+	err := proto.Unmarshal(buf, &hdr)
 	if err != nil {
-		return Header{}, fmt.Errorf("unmarshalling header: %w", err)
+		return nil, fmt.Errorf("unmarshalling header: %w", err)
 	}
 
-	return hdr, nil
+	metricDeviceRecvDecompressedBytes.WithLabelValues(c.idString).Add(float64(2 + len(buf)))
+
+	return &hdr, nil
 }
 
-func (c *rawConnection) handleIndex(im Index) error {
-	l.Debugf("Index(%v, %v, %d file)", c.id, im.Folder, len(im.Files))
-	return c.receiver.Index(c.id, im.Folder, im.Files)
+func (c *rawConnection) handleIndex(im *Index) error {
+	l.Debugf("Index(%v, %v, %d file)", c.deviceID, im.Folder, len(im.Files))
+	return c.model.Index(im)
 }
 
-func (c *rawConnection) handleIndexUpdate(im IndexUpdate) error {
-	l.Debugf("queueing IndexUpdate(%v, %v, %d files)", c.id, im.Folder, len(im.Files))
-	return c.receiver.IndexUpdate(c.id, im.Folder, im.Files)
+func (c *rawConnection) handleIndexUpdate(im *IndexUpdate) error {
+	l.Debugf("queueing IndexUpdate(%v, %v, %d files)", c.deviceID, im.Folder, len(im.Files))
+	return c.model.IndexUpdate(im)
 }
 
 // checkIndexConsistency verifies a number of invariants on FileInfos received in
@@ -652,26 +665,28 @@ func checkFilename(name string) error {
 	return nil
 }
 
-func (c *rawConnection) handleRequest(req Request) {
-	res, err := c.receiver.Request(c.id, req.Folder, req.Name, int32(req.BlockNo), int32(req.Size), req.Offset, req.Hash, req.WeakHash, req.FromTemporary)
+func (c *rawConnection) handleRequest(req *Request) {
+	res, err := c.model.Request(req)
 	if err != nil {
-		c.send(context.Background(), &Response{
+		resp := &Response{
 			ID:   req.ID,
 			Code: errorToCode(err),
-		}, nil)
+		}
+		c.send(context.Background(), resp.toWire(), nil)
 		return
 	}
 	done := make(chan struct{})
-	c.send(context.Background(), &Response{
+	resp := &Response{
 		ID:   req.ID,
 		Data: res.Data(),
 		Code: errorToCode(nil),
-	}, done)
+	}
+	c.send(context.Background(), resp.toWire(), done)
 	<-done
 	res.Close()
 }
 
-func (c *rawConnection) handleResponse(resp Response) {
+func (c *rawConnection) handleResponse(resp *Response) {
 	c.awaitingMut.Lock()
 	if rc := c.awaiting[resp.ID]; rc != nil {
 		delete(c.awaiting, resp.ID)
@@ -681,7 +696,7 @@ func (c *rawConnection) handleResponse(resp Response) {
 	c.awaitingMut.Unlock()
 }
 
-func (c *rawConnection) send(ctx context.Context, msg message, done chan struct{}) bool {
+func (c *rawConnection) send(ctx context.Context, msg proto.Message, done chan struct{}) bool {
 	select {
 	case c.outbox <- asyncMessage{msg, done}:
 		return true
@@ -697,7 +712,7 @@ func (c *rawConnection) send(ctx context.Context, msg message, done chan struct{
 func (c *rawConnection) writerLoop() {
 	select {
 	case cc := <-c.clusterConfigBox:
-		err := c.writeMessage(cc)
+		err := c.writeMessage(cc.toWire())
 		if err != nil {
 			c.internalClose(err)
 			return
@@ -710,9 +725,20 @@ func (c *rawConnection) writerLoop() {
 		return
 	}
 	for {
+		// When the connection is closing or closed, that should happen
+		// immediately, not compete with the (potentially very busy) outbox.
+		select {
+		case hm := <-c.closeBox:
+			_ = c.writeMessage(hm.msg)
+			close(hm.done)
+			return
+		case <-c.closed:
+			return
+		default:
+		}
 		select {
 		case cc := <-c.clusterConfigBox:
-			err := c.writeMessage(cc)
+			err := c.writeMessage(cc.toWire())
 			if err != nil {
 				c.internalClose(err)
 				return
@@ -738,15 +764,19 @@ func (c *rawConnection) writerLoop() {
 	}
 }
 
-func (c *rawConnection) writeMessage(msg message) error {
+func (c *rawConnection) writeMessage(msg proto.Message) error {
 	msgContext, _ := messageContext(msg)
 	l.Debugf("Writing %v", msgContext)
 
-	size := msg.ProtoSize()
-	hdr := Header{
+	defer func() {
+		metricDeviceSentMessages.WithLabelValues(c.idString).Inc()
+	}()
+
+	size := proto.Size(msg)
+	hdr := &bep.Header{
 		Type: typeOf(msg),
 	}
-	hdrSize := hdr.ProtoSize()
+	hdrSize := proto.Size(hdr)
 	if hdrSize > 1<<16-1 {
 		panic("impossibly large header")
 	}
@@ -757,7 +787,7 @@ func (c *rawConnection) writeMessage(msg message) error {
 	defer BufferPool.Put(buf)
 
 	// Message
-	if _, err := msg.MarshalTo(buf[2+hdrSize+4:]); err != nil {
+	if _, err := protoutil.MarshalTo(buf[overhead:], msg); err != nil {
 		return fmt.Errorf("marshalling message: %w", err)
 	}
 
@@ -768,10 +798,12 @@ func (c *rawConnection) writeMessage(msg message) error {
 		}
 	}
 
+	metricDeviceSentUncompressedBytes.WithLabelValues(c.idString).Add(float64(totSize))
+
 	// Header length
 	binary.BigEndian.PutUint16(buf, uint16(hdrSize))
 	// Header
-	if _, err := hdr.MarshalTo(buf[2:]); err != nil {
+	if _, err := protoutil.MarshalTo(buf[2:], hdr); err != nil {
 		return fmt.Errorf("marshalling header: %w", err)
 	}
 	// Message length
@@ -790,17 +822,20 @@ func (c *rawConnection) writeMessage(msg message) error {
 //
 // The first return value indicates whether compression succeeded.
 // If not, the caller should retry without compression.
-func (c *rawConnection) writeCompressedMessage(msg message, marshaled []byte) (ok bool, err error) {
-	hdr := Header{
+func (c *rawConnection) writeCompressedMessage(msg proto.Message, marshaled []byte) (ok bool, err error) {
+	hdr := &bep.Header{
 		Type:        typeOf(msg),
-		Compression: MessageCompressionLZ4,
+		Compression: bep.MessageCompression_MESSAGE_COMPRESSION_LZ4,
 	}
-	hdrSize := hdr.ProtoSize()
+	hdrSize := proto.Size(hdr)
 	if hdrSize > 1<<16-1 {
 		panic("impossibly large header")
 	}
 
 	cOverhead := 2 + hdrSize + 4
+
+	metricDeviceSentUncompressedBytes.WithLabelValues(c.idString).Add(float64(cOverhead + len(marshaled)))
+
 	// The compressed size may be at most n-n/32 = .96875*n bytes,
 	// I.e., if we can't save at least 3.125% bandwidth, we forgo compression.
 	// This number is arbitrary but cheap to compute.
@@ -817,7 +852,7 @@ func (c *rawConnection) writeCompressedMessage(msg message, marshaled []byte) (o
 	// Header length
 	binary.BigEndian.PutUint16(buf, uint16(hdrSize))
 	// Header
-	if _, err := hdr.MarshalTo(buf[2:]); err != nil {
+	if _, err := protoutil.MarshalTo(buf[2:], hdr); err != nil {
 		return true, fmt.Errorf("marshalling header: %w", err)
 	}
 	// Message length
@@ -831,72 +866,72 @@ func (c *rawConnection) writeCompressedMessage(msg message, marshaled []byte) (o
 	return true, nil
 }
 
-func typeOf(msg message) MessageType {
+func typeOf(msg proto.Message) bep.MessageType {
 	switch msg.(type) {
-	case *ClusterConfig:
-		return MessageTypeClusterConfig
-	case *Index:
-		return MessageTypeIndex
-	case *IndexUpdate:
-		return MessageTypeIndexUpdate
-	case *Request:
-		return MessageTypeRequest
-	case *Response:
-		return MessageTypeResponse
-	case *DownloadProgress:
-		return MessageTypeDownloadProgress
-	case *Ping:
-		return MessageTypePing
-	case *Close:
-		return MessageTypeClose
+	case *bep.ClusterConfig:
+		return bep.MessageType_MESSAGE_TYPE_CLUSTER_CONFIG
+	case *bep.Index:
+		return bep.MessageType_MESSAGE_TYPE_INDEX
+	case *bep.IndexUpdate:
+		return bep.MessageType_MESSAGE_TYPE_INDEX_UPDATE
+	case *bep.Request:
+		return bep.MessageType_MESSAGE_TYPE_REQUEST
+	case *bep.Response:
+		return bep.MessageType_MESSAGE_TYPE_RESPONSE
+	case *bep.DownloadProgress:
+		return bep.MessageType_MESSAGE_TYPE_DOWNLOAD_PROGRESS
+	case *bep.Ping:
+		return bep.MessageType_MESSAGE_TYPE_PING
+	case *bep.Close:
+		return bep.MessageType_MESSAGE_TYPE_CLOSE
 	default:
 		panic("bug: unknown message type")
 	}
 }
 
-func newMessage(t MessageType) (message, error) {
+func newMessage(t bep.MessageType) (proto.Message, error) {
 	switch t {
-	case MessageTypeClusterConfig:
-		return new(ClusterConfig), nil
-	case MessageTypeIndex:
-		return new(Index), nil
-	case MessageTypeIndexUpdate:
-		return new(IndexUpdate), nil
-	case MessageTypeRequest:
-		return new(Request), nil
-	case MessageTypeResponse:
-		return new(Response), nil
-	case MessageTypeDownloadProgress:
-		return new(DownloadProgress), nil
-	case MessageTypePing:
-		return new(Ping), nil
-	case MessageTypeClose:
-		return new(Close), nil
+	case bep.MessageType_MESSAGE_TYPE_CLUSTER_CONFIG:
+		return new(bep.ClusterConfig), nil
+	case bep.MessageType_MESSAGE_TYPE_INDEX:
+		return new(bep.Index), nil
+	case bep.MessageType_MESSAGE_TYPE_INDEX_UPDATE:
+		return new(bep.IndexUpdate), nil
+	case bep.MessageType_MESSAGE_TYPE_REQUEST:
+		return new(bep.Request), nil
+	case bep.MessageType_MESSAGE_TYPE_RESPONSE:
+		return new(bep.Response), nil
+	case bep.MessageType_MESSAGE_TYPE_DOWNLOAD_PROGRESS:
+		return new(bep.DownloadProgress), nil
+	case bep.MessageType_MESSAGE_TYPE_PING:
+		return new(bep.Ping), nil
+	case bep.MessageType_MESSAGE_TYPE_CLOSE:
+		return new(bep.Close), nil
 	default:
 		return nil, errUnknownMessage
 	}
 }
 
-func (c *rawConnection) shouldCompressMessage(msg message) bool {
+func (c *rawConnection) shouldCompressMessage(msg proto.Message) bool {
 	switch c.compression {
 	case CompressionNever:
 		return false
 
 	case CompressionAlways:
 		// Use compression for large enough messages
-		return msg.ProtoSize() >= compressionThreshold
+		return proto.Size(msg) >= compressionThreshold
 
 	case CompressionMetadata:
-		_, isResponse := msg.(*Response)
+		_, isResponse := msg.(*bep.Response)
 		// Compress if it's large enough and not a response message
-		return !isResponse && msg.ProtoSize() >= compressionThreshold
+		return !isResponse && proto.Size(msg) >= compressionThreshold
 
 	default:
 		panic("unknown compression setting")
 	}
 }
 
-// Close is called when the connection is regularely closed and thus the Close
+// Close is called when the connection is regularly closed and thus the Close
 // BEP message is sent before terminating the actual connection. The error
 // argument specifies the reason for closing the connection.
 func (c *rawConnection) Close(err error) {
@@ -904,7 +939,7 @@ func (c *rawConnection) Close(err error) {
 		done := make(chan struct{})
 		timeout := time.NewTimer(CloseTimeout)
 		select {
-		case c.closeBox <- asyncMessage{&Close{err.Error()}, done}:
+		case c.closeBox <- asyncMessage{&bep.Close{Reason: err.Error()}, done}:
 			select {
 			case <-done:
 			case <-timeout.C:
@@ -925,9 +960,11 @@ func (c *rawConnection) Close(err error) {
 // internalClose is called if there is an unexpected error during normal operation.
 func (c *rawConnection) internalClose(err error) {
 	c.closeOnce.Do(func() {
-		l.Debugln("close due to", err)
+		c.startStopMut.Lock()
+
+		l.Debugf("close connection to %s at %s due to %v", c.deviceID.Short(), c.ConnectionInfo, err)
 		if cerr := c.closer.Close(); cerr != nil {
-			l.Debugln(c.id, "failed to close underlying conn:", cerr)
+			l.Debugf("failed to close underlying conn %s at %s %v:", c.deviceID.Short(), c.ConnectionInfo, cerr)
 		}
 		close(c.closed)
 
@@ -940,9 +977,17 @@ func (c *rawConnection) internalClose(err error) {
 		}
 		c.awaitingMut.Unlock()
 
-		<-c.dispatcherLoopStopped
+		if !c.startTime.IsZero() {
+			// Wait for the dispatcher loop to exit, if it was started to
+			// begin with.
+			<-c.dispatcherLoopStopped
+		}
 
-		c.receiver.Closed(c.ID(), err)
+		c.startStopMut.Unlock()
+
+		// We don't want to call into the model while holding the
+		// startStopMut.
+		c.model.Closed(err)
 	})
 }
 
@@ -960,11 +1005,11 @@ func (c *rawConnection) pingSender() {
 		case <-ticker.C:
 			d := time.Since(c.cw.Last())
 			if d < PingSendInterval/2 {
-				l.Debugln(c.id, "ping skipped after wr", d)
+				l.Debugln(c.deviceID, "ping skipped after wr", d)
 				continue
 			}
 
-			l.Debugln(c.id, "ping -> after", d)
+			l.Debugln(c.deviceID, "ping -> after", d)
 			c.ping()
 
 		case <-c.closed:
@@ -985,11 +1030,11 @@ func (c *rawConnection) pingReceiver() {
 		case <-ticker.C:
 			d := time.Since(c.cr.Last())
 			if d > ReceiveTimeout {
-				l.Debugln(c.id, "ping timeout", d)
+				l.Debugln(c.deviceID, "ping timeout", d)
 				c.internalClose(ErrTimeout)
 			}
 
-			l.Debugln(c.id, "last read within", d)
+			l.Debugln(c.deviceID, "last read within", d)
 
 		case <-c.closed:
 			return
@@ -1048,25 +1093,57 @@ func newHandleError(err error, msgContext string) error {
 	return fmt.Errorf("handling %v: %w", msgContext, err)
 }
 
-func messageContext(msg message) (string, error) {
+func messageContext(msg proto.Message) (string, error) {
 	switch msg := msg.(type) {
-	case *ClusterConfig:
+	case *bep.ClusterConfig:
 		return "cluster-config", nil
-	case *Index:
+	case *bep.Index:
 		return fmt.Sprintf("index for %v", msg.Folder), nil
-	case *IndexUpdate:
+	case *bep.IndexUpdate:
 		return fmt.Sprintf("index-update for %v", msg.Folder), nil
-	case *Request:
+	case *bep.Request:
 		return fmt.Sprintf(`request for "%v" in %v`, msg.Name, msg.Folder), nil
-	case *Response:
+	case *bep.Response:
 		return "response", nil
-	case *DownloadProgress:
+	case *bep.DownloadProgress:
 		return fmt.Sprintf("download-progress for %v", msg.Folder), nil
-	case *Ping:
+	case *bep.Ping:
 		return "ping", nil
-	case *Close:
+	case *bep.Close:
 		return "close", nil
 	default:
 		return "", errors.New("unknown or empty message")
 	}
 }
+
+// connectionWrappingModel takes the Model interface from the model package,
+// which expects the Connection as the first parameter in all methods, and
+// wraps it to conform to the rawModel interface.
+type connectionWrappingModel struct {
+	conn  Connection
+	model Model
+}
+
+func (c *connectionWrappingModel) Index(m *Index) error {
+	return c.model.Index(c.conn, m)
+}
+
+func (c *connectionWrappingModel) IndexUpdate(idxUp *IndexUpdate) error {
+	return c.model.IndexUpdate(c.conn, idxUp)
+}
+
+func (c *connectionWrappingModel) Request(req *Request) (RequestResponse, error) {
+	return c.model.Request(c.conn, req)
+}
+
+func (c *connectionWrappingModel) ClusterConfig(config *ClusterConfig) error {
+	return c.model.ClusterConfig(c.conn, config)
+}
+
+func (c *connectionWrappingModel) Closed(err error) {
+	c.model.Closed(c.conn, err)
+}
+
+func (c *connectionWrappingModel) DownloadProgress(p *DownloadProgress) error {
+	return c.model.DownloadProgress(c.conn, p)
+}
